(******************************************************************************)
(* generator                                                                  *)
(* (C) 2010..2021 Steffen Solyga <solyga@gmx.de>                              *)
(******************************************************************************)
IMPLEMENTATION MODULE Gen;

IMPORT FIO;
IMPORT Comp;

CONST
  debug = FALSE;
  out ::= FIO.StandardOutput;
  err ::= FIO.ErrorOutput;
  dbg ::= FIO.ErrorOutput;

  acMax = 3;   (* 3 args are enough for l04 (no relations/conditions) *)

TYPE
  TStr = ARRAY [ 0 .. 2 ] OF CHAR;

VAR
  ac: CARDINAL;
  oprs: ARRAY TOperation OF TStr;
  vc: CARDINAL;


PROCEDURE Dbg0();
BEGIN
  FIO.WrStr( dbg, 'Gen.Debug: vars args' );
  FIO.WrLn( dbg );
END Dbg0;

PROCEDURE DbgS( s: ARRAY OF CHAR );
BEGIN
  FIO.WrStr( dbg, 'Gen.Debug:  ' );
  FIO.WrCard( dbg, vc, 2 );
  FIO.WrStr( dbg, '    ' );
  FIO.WrCard( dbg, ac, 1 );
  FIO.WrStr( dbg, ':  ' );
  FIO.WrStr( dbg, s );
END DbgS;

PROCEDURE DbgX();
BEGIN
  FIO.WrLn( dbg );
END DbgX;

PROCEDURE DbgV( v: TValue );
BEGIN
  FIO.WrStr( dbg, '  ' );
  FIO.WrInt( dbg, v, 0 );
  FIO.WrLn( dbg );
END DbgV;

PROCEDURE DbgO( o: TOffset );
BEGIN
  FIO.WrStr( dbg, '   ' );
  FIO.WrCard( dbg, o, 1 );
  FIO.WrLn( dbg );
END DbgO;


PROCEDURE Err( s: ARRAY OF CHAR );
BEGIN
  FIO.WrStr( err, 'Gen: ' );
  FIO.WrStr( err, s );
  FIO.WrStr( err, '.' );
  FIO.WrLn( err );
  HALT;
END Err;

PROCEDURE Value( v: TValue ): Comp.TValue;
BEGIN
  RETURN v;
END Value;

PROCEDURE Index( o: TOffset ): Comp.TIndex;
BEGIN
  INC( o, acMax );  (* acMax vars reserved for calc stack *)
  IF o >= Comp.cnt THEN Err( 'Stack fault' ) END;
  RETURN o;
END Index;


PROCEDURE Val( v: TValue );
BEGIN
  IF debug THEN DbgS( 'Val' ); DbgV( v ) END;
  IF ac >= acMax THEN Err( 'Too many arguments' ) END;
  Comp.Ldv( Value(v) );
  Comp.Sto( ac );
  INC( ac );
END Val;

PROCEDURE Lod( o: TOffset );
BEGIN
  IF debug THEN DbgS( 'Lod' ); DbgO( o ) END;
  IF ac >= acMax THEN Err( 'Too many arguments' ) END;
  Comp.Ldi( Index(o) );
  Comp.Sto( ac );
  INC( ac );
END Lod;

PROCEDURE Dec( x: CARDINAL );
BEGIN
  IF ac < x THEN Err( 'Too little arguments' ) END;
  DEC( ac, x );
END Dec;

PROCEDURE Sto( o: TOffset );
BEGIN
  IF debug THEN DbgS( 'Sto' ); DbgO( o ) END;
  Dec( 1 );
  Comp.Ldi( ac );
  Comp.Sto( Index(o) );
END Sto;

PROCEDURE Calc( o: Comp.TOperation );
BEGIN
  Dec( 2 ); Comp.Ldi( ac ); Comp.Opi( o, ac + 1 );
  Comp.Sto( ac ); INC( ac );
END Calc;

PROCEDURE Opr( o: TOperation );
BEGIN
  IF debug THEN DbgS( oprs[o] ); DbgX() END;
  CASE o OF
  | oNeg: Dec(1); Comp.Ldi(ac); Comp.Neg(); Comp.Sto(ac); INC(ac);
  | oWrt: Dec(1); Comp.Ldi(ac); Comp.Out();
  | oAdd: Calc( Comp.oAdd );
  | oSub: Calc( Comp.oSub );
  | oMul: Calc( Comp.oMul );
  | oDiv: Calc( Comp.oDiv );
  | oNew: INC( vc ); IF vc > Comp.cnt THEN Err( 'Too many variables' ) END;
  ELSE
    Err( 'Bug: Invalid operation' );
  END;
END Opr;

BEGIN
  ac := 0;
  oprs[ oNeg ] := 'Neg';
  oprs[ oWrt ] := 'Wrt';
  oprs[ oAdd ] := 'Add';
  oprs[ oSub ] := 'Sub';
  oprs[ oMul ] := 'Mul';
  oprs[ oDiv ] := 'Div';
  oprs[ oNew ] := 'New';
  vc := acMax;              (* acMax vars reserved for calc stack *)
  IF debug THEN Dbg0() END;
END Gen.
